# lvmpv.py
# Device format classes for anaconda's storage configuration module.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import os

from ..storage_log import log_method_call
from parted import PARTITION_LVM
from ..devicelibs import lvm
from ..tasks import availability, pvtask
from ..i18n import N_
from ..size import Size
from ..errors import PhysicalVolumeError
from . import DeviceFormat, register_device_format
from .. import udev
from ..static_data.lvm_info import pvs_info

import logging
log = logging.getLogger("blivet")


class LVMPhysicalVolume(DeviceFormat):

    """ An LVM physical volume. """
    _type = "lvmpv"
    _name = N_("physical volume (LVM)")
    _udev_types = ["LVM2_member"]
    parted_flag = PARTITION_LVM
    _formattable = True                 # can be formatted
    _supported = True                   # is supported
    _linux_native = True                 # for clearpart
    _min_size = lvm.LVM_PE_SIZE * 2      # one for metadata and one for data
    _packages = ["lvm2"]                # required packages
    _ks_mountpoint = "pv."
    _plugin = availability.BLOCKDEV_LVM_PLUGIN
    _resizable = True

    _size_info_class = pvtask.PVSize
    _resize_class = pvtask.PVResize

    def __init__(self, **kwargs):
        """
            :keyword device: path to the block device node
            :keyword uuid: this PV's uuid (not the VG uuid)
            :keyword exists: indicates whether this is an existing format
            :type exists: bool
            :keyword vg_name: the name of the VG this PV belongs to
            :keyword vg_uuid: the UUID of the VG this PV belongs to
            :keyword pe_start: offset of first physical extent
            :type pe_start: :class:`~.size.Size`
            :keyword data_alignment: data alignment (for non-existent PVs)
            :type data_alignment: :class:`~.size.Size`
            :keyword free: free space in the PV
            :type free: :class:`~.size.Size`

            .. note::

                The 'device' kwarg is required for existing formats. For non-
                existent formats, it is only necessary that the :attr:`device`
                attribute be set before the :meth:`create` method runs. Note
                that you can specify the device at the last moment by specifying
                it via the 'device' kwarg to the :meth:`create` method.
        """
        log_method_call(self, **kwargs)
        DeviceFormat.__init__(self, **kwargs)

        self._size_info = self._size_info_class(self)
        self._resize = self._resize_class(self)

        self.vg_name = kwargs.get("vg_name")
        self.vg_uuid = kwargs.get("vg_uuid")
        self.pe_start = kwargs.get("pe_start", lvm.LVM_PE_START)
        self.data_alignment = kwargs.get("data_alignment", Size(0))
        self._free = kwargs.get("free")  # None means unknown

        self.inconsistent_vg = False

    def __repr__(self):
        s = DeviceFormat.__repr__(self)
        s += ("  vg_name = %(vg_name)s  vg_uuid = %(vg_uuid)s"
              "  pe_start = %(pe_start)s  data_alignment = %(data_alignment)s" %
              {"vg_name": self.vg_name, "vg_uuid": self.vg_uuid,
               "pe_start": self.pe_start, "data_alignment": self.data_alignment})
        return s

    @property
    def dict(self):
        d = super(LVMPhysicalVolume, self).dict
        d.update({"vg_name": self.vg_name, "vg_uuid": self.vg_uuid,
                  "pe_start": self.pe_start, "data_alignment": self.data_alignment})
        return d

    @property
    def formattable(self):
        return super(LVMPhysicalVolume, self).formattable and self._plugin.available

    @property
    def supported(self):
        return super(LVMPhysicalVolume, self).supported and self._plugin.available

    def lvmdevices_add(self):
        if not lvm.HAVE_LVMDEVICES:
            raise PhysicalVolumeError("LVM devices file feature is not supported")

        try:
            blockdev.lvm.devices_add(self.device)
        except blockdev.LVMError as e:
            log.debug("Failed to add PV %s to the LVM devices file: %s", self.device, str(e))

    def lvmdevices_remove(self):
        if not lvm.HAVE_LVMDEVICES:
            raise PhysicalVolumeError("LVM devices file feature is not supported")

        try:
            blockdev.lvm.devices_delete(self.device)
        except blockdev.LVMError as e:
            log.debug("Failed to remove PV %s from the LVM devices file: %s", self.device, str(e))

    def _create(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        lvm.lvm_devices_add(self.device)

        ea_yes = blockdev.ExtraArg.new("-y", "")

        if lvm.HAVE_LVMDEVICES:
            with lvm.empty_lvm_devices():
                # with lvmdbusd we need to call the pvcreate without --devices otherwise lvmdbusd
                # wouldn't be able to find the newly created pv and the call would fail
                blockdev.lvm.pvcreate(self.device, data_alignment=self.data_alignment, extra=[ea_yes])
                self.lvmdevices_add()
        else:
            blockdev.lvm.pvcreate(self.device, data_alignment=self.data_alignment, extra=[ea_yes])

    def _destroy(self, **kwargs):
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        try:
            blockdev.lvm.pvremove(self.device)
        except blockdev.LVMError:
            DeviceFormat._destroy(self, **kwargs)
        finally:
            lvm.lvm_devices_remove(self.device)
            udev.settle()
            if lvm.HAVE_LVMDEVICES:
                self.lvmdevices_remove()

    @property
    def destroyable(self):
        return self._plugin.available

    @property
    def status(self):
        # XXX hack
        return (self.exists and self.vg_name and
                os.path.isdir("/dev/%s" % self.vg_name))

    def update_size_info(self):
        """ Update this format's current size. """

        self._resizable = False

        if not self.exists:
            return

        try:
            self._size = self._size_info.do_task()
        except PhysicalVolumeError as e:
            log.warning("Failed to obtain current size for device %s: %s", self.device, e)
        else:
            self._resizable = True

    @property
    def free(self):
        """ Information about the free space in this PV """
        if self._free is None:
            if self.exists:
                # we don't have any actual value, but the PV exists and is
                # active, we should try to determine it
                pv_info = pvs_info.cache.get(self.device)
                if pv_info is None:
                    log.error("Failed to get free space information for the PV '%s'", self.device)
                    self._free = Size(0)
                else:
                    self._free = Size(pv_info.pv_free)
            else:
                raise PhysicalVolumeError("Unknown free space information for the PV '%s'" % self.device)

        return self._free

    @free.setter
    def free(self, value):
        self._free = value

    @property
    def container_uuid(self):
        return self.vg_uuid

    @container_uuid.setter
    def container_uuid(self, uuid):
        self.vg_uuid = uuid


register_device_format(LVMPhysicalVolume)



#
# btrfs.py
# btrfs functions
#
# Copyright (C) 2011-2014  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): David Lehman <dlehman@redhat.com>
#

from . import raid
from ..size import Size
from ..tasks import availability

import logging
log = logging.getLogger("blivet")

# this is the volume id btrfs always assigns to the top-level volume/tree
MAIN_VOLUME_ID = 5

# if any component device is less than this size, mkfs.btrfs will fail
MIN_MEMBER_SIZE = Size("16 MiB")

raid_levels = raid.RAIDLevels(["raid0", "raid1", "raid10", "single"])

metadata_levels = raid.RAIDLevels(["raid0", "raid1", "raid10", "single", "dup"])

EXTERNAL_DEPENDENCIES = [availability.BLOCKDEV_BTRFS_PLUGIN]


safe_name_characters = "0-9a-zA-Z._@/-"


def is_btrfs_name_valid(name):
    return '\x00' not in name



# populator/helpers/lvm.py
# LVM backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from ...callbacks import callbacks
from ... import udev
from ...devicelibs import lvm
from ...devices.lvm import LVMVolumeGroupDevice, LVMLogicalVolumeDevice, LVMInternalLVtype
from ...errors import DeviceTreeError, DuplicateVGError
from ...flags import flags
from ...size import Size
from ...storage_log import log_method_call
from ...tasks.availability import BLOCKDEV_LVM_PLUGIN_VDO
from .devicepopulator import DevicePopulator
from .formatpopulator import FormatPopulator

from ...static_data import lvs_info, pvs_info, vgs_info

import logging
log = logging.getLogger("blivet")


class LVMDevicePopulator(DevicePopulator):
    @classmethod
    def match(cls, data):
        return udev.device_is_dm_lvm(data)

    def run(self):
        name = udev.device_get_name(self.data)
        log_method_call(self, name=name)

        vg_name = udev.device_get_lv_vg_name(self.data)
        device = self._devicetree.get_device_by_name(vg_name, hidden=True)
        if device and not isinstance(device, LVMVolumeGroupDevice):
            log.warning("found non-vg device with name %s", vg_name)
            device = None

        self._devicetree._add_parent_devices(self.data)

        # LVM provides no means to resolve conflicts caused by duplicated VG
        # names, so we're just being optimistic here. Woo!
        vg_name = udev.device_get_lv_vg_name(self.data)
        vg_device = self._devicetree.get_device_by_name(vg_name)
        if not vg_device:
            log.error("failed to find vg '%s' after scanning pvs", vg_name)

        return self._devicetree.get_device_by_name(name)

    def _handle_rename(self):
        name = self.data.get("DM_LV_NAME")
        if not name:
            return

        # device.name is of the form "%s-%s" % (vg_name, lv_name), while
        # device.lvname is the name of the lv without the vg name.
        if self.device.lvname != name:
            self.device.name = name
        # TODO: update name registry


class LVMFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "lvmpv"

    def _get_kwargs(self):
        kwargs = super(LVMFormatPopulator, self)._get_kwargs()

        # new PV, add it to the LVM devices list and re-run pvs/lvs/vgs
        lvm.lvm_devices_add(self.device.path)
        pvs_info.drop_cache()
        vgs_info.drop_cache()
        lvs_info.drop_cache()

        pv_info = pvs_info.cache.get(self.device.path, None)

        name = udev.device_get_name(self.data)
        if pv_info:
            if pv_info.vg_name:
                kwargs["vg_name"] = pv_info.vg_name
            else:
                log.warning("PV %s has no vg_name", name)
            if pv_info.vg_uuid:
                kwargs["vg_uuid"] = pv_info.vg_uuid
            else:
                log.warning("PV %s has no vg_uuid", name)
            if pv_info.pe_start:
                kwargs["pe_start"] = Size(pv_info.pe_start)
            else:
                log.warning("PV %s has no pe_start", name)
            if pv_info.pv_free:
                kwargs["free"] = Size(pv_info.pv_free)

        return kwargs

    def _get_vg_device(self):
        return self._devicetree.get_device_by_uuid(self.device.format.container_uuid, incomplete=True)

    def _update_lvs(self):
        """ Handle setup of the LVs in the vg_device. """
        log_method_call(self, pv=self.device.name)
        vg_device = self._get_vg_device()
        if vg_device is None:
            # orphan pv
            return

        vg_name = vg_device.name
        lv_info = dict((k, v) for (k, v) in iter(lvs_info.cache.items())
                       if v.vg_name == vg_name)

        for lv_device in vg_device.lvs[:]:
            if lv_device.name not in lv_info:
                log.info("lv %s was removed", lv_device.name)
                self._devicetree.cancel_disk_actions(vg_device.disks)
                self._devicetree.recursive_remove(lv_device, actions=False)

        if not vg_device.complete:
            log.warning("Skipping LVs for incomplete VG %s", vg_name)
            return

        if not lv_info:
            log.debug("no LVs listed for VG %s", vg_name)
            return

        all_lvs = []
        internal_lvs = []

        def add_required_lv(name, msg):
            """ Add a prerequisite/parent LV.

                The parent is strictly required in order to be able to add
                some other LV that depends on it. For this reason, failure to
                add the specified LV results in a DeviceTreeError with the
                message string specified in the msg parameter.

                :param str name: the full name of the LV (including vgname)
                :param str msg: message to pass DeviceTreeError ctor on error
                :returns: None
                :raises: :class:`~.errors.DeviceTreeError` on failure

            """
            vol = self._devicetree.get_device_by_name(name)
            if vol is None:
                new_lv = add_lv(lv_info[name])
                if new_lv:
                    all_lvs.append(new_lv)
                vol = self._devicetree.get_device_by_name(name)

                if vol is None:
                    log.error("%s: %s", msg, name)
                    raise DeviceTreeError(msg)

        def add_lv(lv):
            """ Instantiate and add an LV based on data from the VG. """
            lv_name = lv.lv_name
            lv_uuid = lv.uuid
            lv_attr = lv.attr
            lv_size = Size(lv.size)
            lv_type = lv.segtype

            lv_parents = [vg_device]
            lv_kwargs = {}
            name = "%s-%s" % (vg_name, lv_name)

            lv_device = self._devicetree.get_device_by_name(name)
            if lv_device is not None:
                # some lvs may have been added on demand below
                log.debug("already added %s", name)
                if lv_size != lv_device.current_size:
                    # lvresize can operate on an inactive lv, in which case
                    # the only notification we will receive is a change uevent
                    # for the pv(s)
                    old_size = lv_device._size
                    lv_device.update_size(newsize=lv_size)
                    callbacks.attribute_changed(device=lv_device, attr="size",
                                                old=old_size, new=lv_size)
                    self._devicetree.cancel_disk_actions(vg_device.disks)

                return

            if lv_attr[0] in 'Ss':
                log.info("found lvm snapshot volume '%s'", name)
                origin_name = blockdev.lvm.lvorigin(vg_name, lv_name)
                if not origin_name:
                    log.error("lvm snapshot '%s-%s' has unknown origin",
                              vg_name, lv_name)
                    return

                if origin_name.endswith("_vorigin]"):
                    lv_kwargs["vorigin"] = True
                    origin = None
                else:
                    origin_device_name = "%s-%s" % (vg_name, origin_name)
                    add_required_lv(origin_device_name,
                                    "failed to locate origin lv")
                    origin = self._devicetree.get_device_by_name(origin_device_name)

                lv_kwargs["origin"] = origin
            elif lv_attr[0] in 'IrielTCoD' and lv_name.endswith(']'):
                # an internal LV, add the an instance of the appropriate class
                # to internal_lvs for later processing when non-internal LVs are
                # processed
                internal_lvs.append(lv_name)
                return
            elif lv_attr[0] == 't':
                # thin pool
                # nothing to do here
                pass
            elif lv_attr[0] == 'V':
                # thin volume
                pool_name = blockdev.lvm.thlvpoolname(vg_name, lv_name)
                pool_device_name = "%s-%s" % (vg_name, pool_name)
                add_required_lv(pool_device_name, "failed to look up thin pool")

                origin_name = blockdev.lvm.lvorigin(vg_name, lv_name)
                if origin_name:
                    origin_device_name = "%s-%s" % (vg_name, origin_name)
                    add_required_lv(origin_device_name, "failed to locate origin lv")
                    origin = self._devicetree.get_device_by_name(origin_device_name)
                    lv_kwargs["origin"] = origin

                lv_parents = [self._devicetree.get_device_by_name(pool_device_name)]
            elif lv_attr[0] == 'd':
                # vdo pool
                if BLOCKDEV_LVM_PLUGIN_VDO.available:
                    pool_info = blockdev.lvm.vdo_info(vg_name, lv_name)
                    if pool_info:
                        lv_kwargs["compression"] = pool_info.compression
                        lv_kwargs["deduplication"] = pool_info.deduplication
                    else:
                        log.warning("failed to get information about VDO pool %s", lv_name)
            elif lv_attr[0] == 'v':
                if lv_type != "vdo":
                    # skip vorigins
                    return
                pool_name = blockdev.lvm.vdolvpoolname(vg_name, lv_name)
                pool_device_name = "%s-%s" % (vg_name, pool_name)
                add_required_lv(pool_device_name, "failed to look up VDO pool")

                lv_parents = [self._devicetree.get_device_by_name(pool_device_name)]
            elif lv_name.endswith(']'):
                # unrecognized Internal LVM2 device
                return
            elif lv_attr[0] not in '-mMrRoOC':
                # Ignore anything else except for the following:
                #   - normal lv
                #   m mirrored
                #   M mirrored without initial sync
                #   r raid
                #   R raid without initial sync
                #   o origin
                #   O origin with merging snapshot
                #   C cached LV
                return

            lv_dev = self._devicetree.get_device_by_uuid(lv_uuid)
            if lv_dev is None:
                lv_device = LVMLogicalVolumeDevice(lv_name, parents=lv_parents,
                                                   uuid=lv_uuid, size=lv_size, seg_type=lv_type,
                                                   exists=True, **lv_kwargs)
                self._devicetree._add_device(lv_device)
                if flags.auto_dev_updates:
                    try:
                        lv_device.setup()
                    except blockdev.LVMError:
                        log.warning("failed to activate lv %s", lv_device.name)
                        lv_device.controllable = False

                if lv_device.status:
                    lv_device.update_sysfs_path()
                    lv_device.update_size()
                    lv_info = udev.get_device(lv_device.sysfs_path)
                    if not lv_info:
                        log.error("failed to get udev data for lv %s", lv_device.name)
                        return lv_device

                    # do format handling now
                    self._devicetree.handle_device(lv_info, update_orig_fmt=True)

                return lv_device

            return None

        def create_internal_lv(lv):
            lv_name = lv.lv_name
            lv_uuid = lv.uuid
            lv_attr = lv.attr
            lv_size = Size(lv.size)
            seg_type = lv.segtype

            lv_type = LVMInternalLVtype.get_type(lv_attr, lv_name)
            if lv_type is LVMInternalLVtype.unknown:
                raise DeviceTreeError("Internal LVs of type '%s' are not supported" % lv_attr[0])

            # strip the "[]"s marking the LV as internal
            lv_name = lv_name.strip("[]")

            # don't know the parent LV yet, will be set later
            new_lv = LVMLogicalVolumeDevice(lv_name, vg_device, parent_lv=None, int_type=lv_type,
                                            size=lv_size, uuid=lv_uuid, exists=True, seg_type=seg_type)
            if new_lv.status:
                new_lv.update_sysfs_path()
                new_lv.update_size()

                lv_info = udev.get_device(new_lv.sysfs_path)
                if not lv_info:
                    log.error("failed to get udev data for lv %s", new_lv.name)
                    return new_lv

            return new_lv

        # add LVs
        for lv in lv_info.values():
            # add the LV to the DeviceTree
            new_lv = add_lv(lv)

            if new_lv:
                # save the reference for later use
                all_lvs.append(new_lv)

        # Instead of doing a topological sort on internal LVs to make sure the
        # parent LV is always created before its internal LVs (an internal LV
        # can have internal LVs), we just create all the instances here and
        # assign their parents later. Those who are not assinged a parent (which
        # would hold a reference to them) will get eaten by the garbage
        # collector.

        # create device instances for the internal LVs
        orphan_lvs = dict()
        for lv_name in internal_lvs:
            full_name = "%s-%s" % (vg_name, lv_name)
            try:
                new_lv = create_internal_lv(lv_info[full_name])
            except DeviceTreeError as e:
                log.warning("Failed to process an internal LV '%s': %s", full_name, e)
            else:
                orphan_lvs[full_name] = new_lv
                all_lvs.append(new_lv)

        # assign parents to internal LVs (and vice versa)
        for lv in orphan_lvs.values():
            parent_lv = lvm.determine_parent_lv(lv, all_lvs, lv_info)
            if parent_lv:
                lv.parent_lv = parent_lv
            else:
                log.warning("Failed to determine parent LV for an internal LV '%s'", lv.name)

    def _add_vg_device(self):
        pv_info = pvs_info.cache.get(self.device.path, None)
        if pv_info:
            vg_name = pv_info.vg_name
            vg_uuid = pv_info.vg_uuid
        else:
            # no info about the PV -> we're done
            return

        if not vg_name:
            log.info("lvm pv %s has no vg", self.device.name)
            return

        vg_device = self._devicetree.get_device_by_uuid(vg_uuid, incomplete=True)
        if vg_device and self.device not in vg_device.parents:
            vg_device.parents.append(self.device)
            callbacks.parent_added(device=vg_device, parent=self.device)
        elif vg_device is None:
            same_name = self._devicetree.get_device_by_name(vg_name)
            if isinstance(same_name, LVMVolumeGroupDevice):
                raise DuplicateVGError("multiple LVM volume groups with the same name (%s)" % vg_name)

            try:
                vg_size = Size(pv_info.vg_size)
                vg_free = Size(pv_info.vg_free)
                pe_size = Size(pv_info.vg_extent_size)
                pe_count = pv_info.vg_extent_count
                pe_free = pv_info.vg_free_count
                pv_count = pv_info.vg_pv_count
            except (KeyError, ValueError) as e:
                log.warning("invalid data for %s: %s", self.device.name, e)
                return

            vg_info = vgs_info.cache.get(pv_info.vg_uuid)
            if vg_info is None:
                log.warning("Failed to get information about LVM volume group %s (%s)", vg_name, pv_info.vg_uuid)
                return

            if not hasattr(vg_info, "exported"):
                log.info("Can't get exported information from VG info, assuming VG is not exported.")
                exported = False
            else:
                exported = vg_info.exported

            vg_device = LVMVolumeGroupDevice(vg_name,
                                             parents=[self.device],
                                             uuid=vg_uuid,
                                             size=vg_size,
                                             free=vg_free,
                                             pe_size=pe_size,
                                             pe_count=pe_count,
                                             pe_free=pe_free,
                                             pv_count=pv_count,
                                             exists=True,
                                             exported=exported)
            self._devicetree._add_device(vg_device)

    def run(self):
        log_method_call(self, pv=self.device.name)
        super(LVMFormatPopulator, self).run()
        self._add_vg_device()
        self._update_lvs()

    def _handle_vg_rename(self):
        vg_device = self._get_vg_device()
        if vg_device is None:
            return

        pv_info = pvs_info.cache.get(self.device.path, None)
        if not pv_info or not pv_info.vg_name:
            return

        vg_name = pv_info.vg_name
        if vg_device.name != vg_name:
            vg_device.name = vg_name
        # TODO: update name registry

    def _update_pv_format(self):
        pv_info = pvs_info.cache.get(self.device.path, None)
        if not pv_info:
            return

        self.device.format.vg_name = pv_info.vg_name
        self.device.format.vg_uuid = pv_info.vg_uuid
        self.device.format.pe_start = Size(pv_info.pe_start)
        self.device.format.pe_free = Size(pv_info.pv_free)

    def update(self):
        self._devicetree.drop_lvm_cache()
        self._update_pv_format()
        pv_info = pvs_info.cache.get(self.device.path, None)
        vg_device = self._get_vg_device()
        if vg_device is None:
            # The VG device isn't in the tree. The PV might have just been
            # added to a VG.
            if pv_info and pv_info.vg_name:
                # Handle adding orphan pv to a vg.
                self._add_vg_device()
            elif self.device.children:
                # The PV was removed from its VG or the VG was removed.
                vg_device = self.device.children[0]
                if len(vg_device.parents) > 1:
                    vg_device.parents.remove(self.device)
                    callbacks.parent_removed(device=vg_device, parent=self.device)
                else:
                    self._devicetree.recursive_remove(vg_device, actions=False)
                return
        else:
            # The VG device is in the tree. Check if the PV still belongs to it.
            if pv_info and pv_info.vg_name:
                # This is the "normal" case: the pv is still part of the
                # vg it was part of last time we looked.
                self._handle_vg_rename()

        # handles vg rename, lv add, lv remove, lv resize (inactive lv)
        self._update_lvs()



# devices/raid.py
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>
#

import abc

from six import add_metaclass

from .. import errors
from ..i18n import _, P_
from ..threads import SynchronizedABCMeta

from .storage import StorageDevice


@add_metaclass(SynchronizedABCMeta)
class RaidDevice(StorageDevice):

    """ Metaclass for devices that support RAID in some form. """
    members = abc.abstractproperty(lambda s: [],
                                   doc="A list of the member device instances")

    def _validate_raid_level(self, level, parent_diff=0):
        """ Returns an error message if the RAID level is invalid for this
            device, otherwise None.

            :param level: a RAID level
            :type level: :class:`~.devicelibs.raid.RAIDLevel`
            :param int parent_diff: number of parents being removed or added
            :returns: an error message if the RAID level is invalid, else None
            :rtype: str or NoneType

            The default number of parents being removed or added is 0,
            indicating that the level is to be checked against the current
            number of parents. The number is positive for added parents,
            negative for removed parents.
        """
        num_members = len(self.members) + parent_diff  # pylint: disable=no-member
        if not self.exists and num_members < level.min_members:
            message = P_(
                "RAID level %(raid_level)s requires that device have at least %(min_members)d member.",
                "RAID level %(raid_level)s requires that device have at least %(min_members)d members.",
                level.min_members
            )
            return message % {"raid_level": level, "min_members": level.min_members}
        return None

    def _get_level(self, value, levels):
        """ Obtains a valid level for the allowed set of levels.

            :param value: a RAID level
            :type value: a valid RAID level descriptor
            :param levels: a list of valid levels
            :type levels: :class:`~.devicelibs.raid.RAIDLevels`
            :returns: a valid RAID level
            :rtype: :class:`~.devicelibs.raid.RAIDLevel`
            :raises ValueError: if invalid RAID level
        """
        try:
            level = levels.raid_level(value)
        except errors.RaidError:
            message = _("RAID level %(raid_level)s is an invalid value. Must be one of (%(levels)s).")
            choices = ", ".join([str(l) for l in levels])
            raise ValueError(message % {"raid_level": value, "levels": choices})

        error_msg = self._validate_raid_level(level)
        if error_msg:
            raise ValueError(error_msg)

        return level

    def _validate_parent_removal(self, level, parent):
        """ Check if it is possible to remove a parent from this device.

            :param level: a RAID level
            :type level: :class:`~.devicelibs.raid.RAIDLevel`
            :param parent: the parent to be removed
            :type parent: :class:`~.devices.StorageDevice`
            :returns: An error message if there is a problem
            :rtype: str or None
        """
        # If this is a raid array that is not actually redundant and it
        # appears to have formatting and therefore probably data on it,
        # removing one of its devices is a bad idea.
        try:
            if not level.has_redundancy() and self.exists and parent.format.exists:
                return _("Cannot remove a member from existing %s array") % level
        except errors.RaidError:
            # If the concept of redundancy is meaningless for this device's
            # raid level, then it is OK to remove a parent device.
            pass

        # Removing a member may invalidate the RAID level
        return self._validate_raid_level(level, -1)



#
# Copyright (C) 2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

import os
import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")

from ..tasks import availability


class MpathMembers(object):
    """A cache for querying multipath member devices"""

    def __init__(self):
        self._members = None

    def is_mpath_member(self, device):
        """Checks if the given device is a member of some multipath mapping or not.

        :param str device: path of the device to query

        """
        if self._members is None:
            if availability.BLOCKDEV_MPATH_PLUGIN.available:
                self._members = set(blockdev.mpath.get_mpath_members())
            else:
                self._members = set()

        device = os.path.realpath(device)
        device = device[len("/dev/"):]

        return device in self._members

    def update_cache(self, device):
        """Update the cache with the given device (checks and adds it is an mpath member)

        :param str device: path of the device to check and add

        """
        device = os.path.realpath(device)
        device = device[len("/dev/"):]

        if availability.BLOCKDEV_MPATH_PLUGIN.available and blockdev.mpath.is_mpath_member(device):
            self._members.add(device)

    def drop_cache(self):
        self._members = None


mpath_members = MpathMembers()



#
# fcoe.py - fcoe class
#
# Copyright (C) 2009  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from . import errors
from . import udev
from . import util
import logging
import time
import shutil
from .i18n import _
log = logging.getLogger("blivet")

_fcoe_module_loaded = False


def has_fcoe():
    global _fcoe_module_loaded
    if not _fcoe_module_loaded:
        util.run_program(["modprobe", "libfc"])
        _fcoe_module_loaded = True
        if "bnx2x" in util.lsmod():
            log.info("fcoe: loading bnx2fc")
            util.run_program(["modprobe", "bnx2fc"])

    return os.access("/sys/module/libfc", os.X_OK)


class FCoE(object):

    """ FCoE utility class.

        This class will automatically discover and connect to EDD configured
        FCoE SAN's when the startup() method gets called. It can also be
        used to manually configure FCoE SAN's through the add_san() method.

        As this class needs to make sure certain things like starting fcoe
        daemons and connecting to firmware discovered SAN's only happens once
        and as it keeps a global list of all FCoE devices it is
        implemented as a Singleton.

        .. warning::
            Since this is a singleton class, calling deepcopy() on the instance
            just returns ``self`` with no copy being created.
    """

    def __init__(self):
        self.started = False
        self.lldpad_started = False
        self.nics = []
        self.added_nics = []

    # So that users can write fcoe() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):
        # pylint: disable=unused-argument
        return self

    def _stabilize(self):
        # I have no clue how long we need to wait, this ought to do the trick
        time.sleep(10)
        udev.settle()

    def _start_edd(self):
        try:
            buf = util.capture_output(["/usr/libexec/fcoe/fcoe_edd.sh", "-i"])
        except OSError as e:
            log.info("Failed to read FCoE EDD info: %s", e.strerror)
            return

        (key, _equals, val) = buf.strip().partition("=")
        if not val or key != "NIC":
            log.info("No FCoE EDD info found: %s", buf.rstrip())
            return

        dcb = self._iface_driver(val) not in ("bnx2x", "bnx2fc")

        log.info("FCoE NIC found in EDD: %s, using dcb: %s", val, dcb)
        self.add_san(val, dcb=dcb, auto_vlan=True)

    def startup(self):
        if self.started:
            return

        if not has_fcoe():
            return

        self._start_edd()
        self.started = True

    def _start_lldpad(self):
        if self.lldpad_started:
            return

        util.run_program(["systemctl", "start", "lldpad.service"])
        self.lldpad_started = True

    def add_san(self, nic, dcb=False, auto_vlan=True):
        """Activates FCoE SANs attached to interface specified by nic.

           Returns error message, or empty string if succeeded.
        """
        if not has_fcoe():
            raise errors.FCoEError(_("FCoE not available"))

        log.info("Activating FCoE SAN attached to %s, dcb: %s autovlan: %s",
                 nic, dcb, auto_vlan)

        rc = 0
        out = ""
        error_msg = ""
        if dcb:
            timeout = 60
            timeout_msg = ""
            self._start_lldpad()

            command_list = [
                ("waiting for lldpad to be ready", ["lldptool", "-p"]),
                ("retrying to turn dcb on", ["dcbtool", "sc", nic, "dcb", "on"]),
                ("retrying to set up dcb with pfc", ["dcbtool", "sc", nic, "pfc", "e:1", "a:1", "w:1"]),
                ("retrying to set up dcb for fcoe", ["dcbtool", "sc", nic, "app:fcoe", "e:1", "a:1", "w:1"]),
            ]

            for timeout_msg, cmd in command_list:
                rc, out = util.run_program_and_capture_output(cmd)
                while rc != 0:
                    timeout -= 1
                    time.sleep(1)
                    if timeout <= 0:
                        break
                    rc, out = util.run_program_and_capture_output(cmd)
                if timeout <= 0:
                    break

            if rc == 0:
                time.sleep(1)
            else:
                log.info("Timed out when %s", timeout_msg)

        if rc == 0:
            self.write_nic_fcoe_cfg(nic, dcb=dcb, auto_vlan=auto_vlan)
            rc, out = util.run_program_and_capture_output(
                ["systemctl", "restart", "fcoe.service"])

        if rc == 0:
            self._stabilize()
            self.nics.append((nic, dcb, auto_vlan))
        else:
            log.debug("Activating FCoE SAN failed: %s %s", rc, out)
            error_msg = out

        return error_msg

    def _iface_driver(self, nic):
        try:
            dpath = os.readlink("/sys/class/net/%s/device/driver" % nic)
        except OSError as e:
            log.debug("Can't find driver of device %s, %s", nic, e)
            driver = ""
        else:
            driver = os.path.basename(dpath)
        return driver

    def write(self, root):
        if not self.nics:
            return

        # Done before packages are installed so don't call
        # write_nic_fcoe_cfg in target root but just copy the cfgs
        dest_cfg_dir = root + "/etc/fcoe"
        # The directory may already exist on image installation (eg RHEV).
        if os.path.isdir(dest_cfg_dir):
            shutil.rmtree(dest_cfg_dir)
        shutil.copytree("/etc/fcoe", dest_cfg_dir)

    def write_nic_fcoe_cfg(self, nic, dcb=True, auto_vlan=True, enable=True, mode=None, root=""):
        cfg_dir = root + "/etc/fcoe"
        example_cfg = os.path.join(cfg_dir, "cfg-ethx")
        if os.access(example_cfg, os.R_OK):
            lines = open(example_cfg, "r").readlines()
        else:
            anaconda_cfg = """FCOE_ENABLE="yes"
DCB_REQUIRED="yes"
AUTO_VLAN="yes"
MODE="fabric"
"""
            lines = anaconda_cfg.splitlines(True)

        with open(os.path.join(cfg_dir, "cfg-%s" % nic), "w") as new_cfg:
            new_cfg.write("# Generated by Anaconda installer\n")
            for line in lines:
                if not line.strip().startswith("#"):
                    if line.startswith("FCOE_ENABLE"):
                        if enable:
                            line = 'FCOE_ENABLE="yes"\n'
                        else:
                            line = 'FCOE_ENABLE="no"\n'
                    elif line.startswith("DCB_REQUIRED"):
                        if dcb:
                            line = 'DCB_REQUIRED="yes"\n'
                        else:
                            line = 'DCB_REQUIRED="no"\n'
                    elif line.startswith("AUTO_VLAN"):
                        if auto_vlan:
                            line = 'AUTO_VLAN="yes"\n'
                        else:
                            line = 'AUTO_VLAN="no"\n'
                    elif line.startswith("MODE"):
                        if mode:
                            line = 'MODE="%s"\n' % mode
                new_cfg.write(line)


# Create FCoE singleton
fcoe = FCoE()

# vim:tw=78:ts=4:et:sw=4



# biosboot.py
# Device format classes for anaconda's storage configuration module.
#
# Copyright (C) 2011  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

from parted import PARTITION_BIOS_GRUB

from ..size import Size
from .. import arch
from ..i18n import N_
from . import DeviceFormat, register_device_format


class BIOSBoot(DeviceFormat):

    """ BIOS boot partition for GPT disklabels. """
    _type = "biosboot"
    _name = N_("BIOS Boot")
    parted_flag = PARTITION_BIOS_GRUB
    _formattable = True                 # can be formatted
    _linux_native = True                 # for clearpart
    _max_size = Size("2 MiB")
    _min_size = Size("512 KiB")
    _supported = True

    def __init__(self, **kwargs):
        """
            :keyword device: path to the block device node
            :type device: str
            :keyword exists: whether the formatting exists
            :type exists: bool

            .. note::

                The 'device' kwarg is required for existing formats.
        """
        DeviceFormat.__init__(self, **kwargs)

    @property
    def status(self):
        return False

    @property
    def supported(self):
        return super(BIOSBoot, self).supported and arch.is_x86() and not arch.is_efi()


register_device_format(BIOSBoot)



#
# raid.py
# representation of RAID levels
#
# Copyright (C) 2009  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): Dave Lehman <dlehman@redhat.com>
#            Anne Mulhern <amulhern@redhat.com>
#

import abc

from six import add_metaclass

from ..errors import RaidError
from ..size import Size


def div_up(a, b):
    """Rounds up integer division.  For example, div_up(3, 2) is 2.

       :param int a: the dividend
       :param int b: the divisor
    """
    return (a + (b - 1)) // b


@add_metaclass(abc.ABCMeta)
class RAIDLevel(object):

    """An abstract class which is the parent of all classes which represent
       a RAID level.

       It ensures that RAIDLevel objects will really be singleton objects
       by overriding copy methods.
    """

    name = abc.abstractproperty(doc="The canonical name for this level")
    names = abc.abstractproperty(doc="List of recognized names for this level.")
    min_members = abc.abstractproperty(doc="The minimum number of members required to make a fully functioning array.")

    @abc.abstractmethod
    def has_redundancy(self):
        """ Whether this RAID level incorporates inherent redundancy.

            Note that for some RAID levels, the notion of redundancy is
            meaningless.

            :rtype: boolean
            :returns: True if this RAID level has inherent redundancy
        """
        raise NotImplementedError()

    is_uniform = abc.abstractproperty(doc="Whether data is uniformly distributed across all devices.")

    def __str__(self):
        return self.name

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        # pylint: disable=unused-argument
        return self


@add_metaclass(abc.ABCMeta)
class RAIDn(RAIDLevel):

    """An abstract class which is the parent of classes which represent a
       numeric RAID level. A better word would be classification, since 'level'
       implies an ordering, but level is the canonical word.

       The abstract properties of the class are:

       - level: A string containing the number that designates this level

       - nick: A single nickname for this level, may be None

        All methods in this class fall into these categories:

        1) May not be overrridden in any subclass.

        2) Are private abstract methods.

        3) Are special Python methods, e.g., __str__


        Note that each subclass in this file is instantiated immediately after
        it is defined and using the same name, effectively yielding a
        singleton object of the class.
    """

    # ABSTRACT PROPERTIES
    level = abc.abstractproperty(doc="A code representing the level")
    nick = abc.abstractproperty(doc="A nickname for this level")

    # PROPERTIES
    is_uniform = property(lambda s: True)

    number = property(lambda s: int(s.level),
                      doc="A numeric code for this level")

    name = property(lambda s: "raid" + s.level,
                    doc="The canonical name for this level")

    alt_synth_names = property(lambda s: ["RAID" + s.level, s.level, s.number],
                               doc="names that can be synthesized from level but are not name")

    names = property(lambda s:
                     [n for n in [s.name] + [s.nick] + s.alt_synth_names if n is not None],
                     doc="all valid names for this level")

    # METHODS
    def get_max_spares(self, member_count):
        """The maximum number of spares for this level.

           :param int member_count: the number of members belonging to the array
           :rtype: int

           Raiess a RaidError if member_count is fewer than the minimum
           number of members required for this level.
        """
        if member_count < self.min_members:
            raise RaidError("%s requires at least %d disks" % (self.name, self.min_members))
        return self._get_max_spares(member_count)

    @abc.abstractmethod
    def _get_max_spares(self, member_count):
        """Helper function; not to be called directly."""
        raise NotImplementedError()

    def get_base_member_size(self, size, member_count):
        """The required size for each member of the array for
           storing only data.

           :param size: size of data to be stored
           :type size: :class:`~.size.Size`
           :param int member_count: number of members in this array
           :rtype: :class:`~.size.Size`

           Raises a RaidError if member_count is fewer than the minimum
           number of members required for this array or if size is less
           than 0.
        """
        if member_count < self.min_members:
            raise RaidError("%s requires at least %d disks" % (self.name, self.min_members))
        if size < 0:
            raise RaidError("size is a negative number")
        return self._get_base_member_size(size, member_count)

    @abc.abstractmethod
    def _get_base_member_size(self, size, member_count):
        """Helper function; not to be called directly."""
        raise NotImplementedError()

    def get_net_array_size(self, member_count, smallest_member_size):
        """Return the space, essentially the number of bits available
           for storage. This value is generally a function of the
           smallest member size. If the smallest member size represents
           the amount of data that can be stored on the smallest member,
           then the result will represent the amount of data that can be
           stored on the array. If the smallest member size represents
           both data and metadata, then the result will represent the
           available space in the array for both data and metadata.

           :param int member_count: the number of members in the array
           :param smallest_member_size: the size of the smallest
             member of this array
           :type smallest_member_size: :class:`~.size.Size`
           :returns: the array size
           :rtype: :class:`~.size.Size`

           Raises a RaidError if member_count is fewer than the minimum
           number of members required for this array or if size is less
           than 0.
        """
        if member_count < self.min_members:
            raise RaidError("%s requires at least %d disks" % (self.name, self.min_members))
        if smallest_member_size < Size(0):
            raise RaidError("size is a negative number")
        return self._get_net_array_size(member_count, smallest_member_size)

    @abc.abstractmethod
    def _get_net_array_size(self, member_count, smallest_member_size):
        """Helper function; not to be called directly."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _trim(self, size, chunk_size):
        """Helper function; not to be called directly.

           Trims size to the largest size that the level allows based on the
           chunk_size.

           :param size: the size of the array
           :type size: :class:`~.size.Size`
           :param chunk_size: the smallest unit of size this array allows
           :type chunk_size: :class:`~.size.Size`
           :rtype: :class:`~.size.Size`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _pad(self, size, chunk_size):
        """Helper function; not to be called directly.

           Pads size to the smallest size greater than size that is in units
           of chunk_size.

           :param size: the size of the array
           :type size: :class:`~.size.Size`
           :param chunk_size: the smallest unit of size this array allows
           :type chunk_size: :class:`~.size.Size`
           :rtype: :class:`~.size.Size`
        """
        raise NotImplementedError()

    def get_recommended_stride(self, member_count):
        """Return a recommended stride size in blocks.

           Returns None if there is no recommended size.

           :param int member_count: the number of members in the array
           :rtype: int or None

           Raises a RaidError if member_count is fewer than the
           minimum number of members required for this level
        """
        if member_count < self.min_members:
            raise RaidError("%s requires at least %d disks" % (self.name, self.min_members))
        return self._get_recommended_stride(member_count)

    @abc.abstractmethod
    def _get_recommended_stride(self, member_count):
        """Helper function; not to be called directly."""
        raise NotImplementedError()

    def get_size(self, member_sizes, num_members=None, chunk_size=None, superblock_size_func=None):
        """Estimate the amount of data that can be stored on this array.

           :param member_size: a list of the sizes of members of this array
           :type member_size: list of :class:`~.size.Size`
           :param int num_members: the number of members in the array
           :param chunk_size: the smallest unit of size read or written
           :type chunk_size: :class:`~.size.Size`
           :param superblock_size_func: a function that estimates the
              superblock size for this array
           :type superblock_size_func: a function from :class:`~.size.Size` to
              :class:`~.size.Size`
           :returns: an estimate of the amount of data that can be stored on
              this array
           :rtype: :class:`~.size.Size`

           Note that the number of members in the array may not be the same
           as the length of member_sizes if the array is still
           under construction.
        """
        if not member_sizes:
            return Size(0)

        if num_members is None:
            num_members = len(member_sizes)

        if chunk_size is None or chunk_size == Size(0):
            raise RaidError("chunk_size parameter value %s is not acceptable")

        if superblock_size_func is None:
            raise RaidError("superblock_size_func value of None is not acceptable")

        min_size = min(member_sizes)
        superblock_size = superblock_size_func(min_size)
        min_data_size = self._trim(min_size - superblock_size, chunk_size)
        return self.get_net_array_size(num_members, min_data_size)

    def get_space(self, size, num_members, chunk_size=None, superblock_size_func=None):
        """Estimate the amount of memory required by this array, including
           memory allocated for metadata.

           :param size: the amount of data on this array
           :type size: :class:`~.size.Size`
           :param int num_members: the number of members in the array
           :param chunk_size: the smallest unit of size read or written
           :type chunk_size: :class:`~.size.Size`
           :param superblock_size_func: a function that estimates the
              superblock size for this array
           :type superblock_size_func: a function from :class:`~.size.Size` to
              :class:`~.size.Size`
           :returns: an estimate of the memory required, including metadata
           :rtype: :class:`~.size.Size`
        """
        if superblock_size_func is None:
            raise RaidError("superblock_size_func value of None is not acceptable")

        size_per_member = self.get_base_member_size(size, num_members)
        size_per_member += superblock_size_func(size)
        if chunk_size is not None:
            size_per_member = self._pad(size_per_member, chunk_size)
        return size_per_member * num_members


class RAIDLevels(object):

    """A class which keeps track of registered RAID levels. This class
       may be extended, overriding the is_raid method to include any
       additional properties that a client of this package may require
       for its RAID levels.
    """

    def __init__(self, levels=None):
        """Add the specified standard levels to the levels in this object.

           :param levels: the levels to be added to this object
           :type levels: list of valid RAID level descriptors

           If levels is True, add all standard levels. Else, levels
           must be a list of valid level descriptors of standard levels.
           Duplicate descriptors are ignored.
        """
        levels = levels or []
        self._raid_levels = set()
        for level in levels:
            matches = [l for l in ALL_LEVELS if level in l.names]
            if len(matches) != 1:
                raise RaidError("invalid standard RAID level descriptor %s" % level)
            else:
                self.add_raid_level(matches[0])

    @classmethod
    def is_raid_level(cls, level):
        """Return False if level does not satisfy minimum requirements for
           a RAID level, otherwise return True.

           :param object level: an object representing a RAID level

           There must be at least one element in the names list, or the level
           will be impossible to look up by any string.

           The name property must be defined; it should be one of the
           elements in the names list.

           All RAID objects that extend RAIDlevel are guaranteed to pass these
           minimum requirements.

           This method should not be overridden in any subclass so that it
           is so restrictive that a RAIDlevel object does not satisfy it.
        """
        return len(level.names) > 0 and level.name in level.names

    def raid_level(self, descriptor):
        """Return RAID object corresponding to descriptor.

           :param object descriptor: a RAID level descriptor

           Note that descriptor may be any object that identifies a
           RAID level, including the RAID object itself.

           Raises a RaidError if no RAID object can be found for this
           descriptor.
        """
        for level in self._raid_levels:
            if descriptor in level.names or descriptor is level:
                return level
        raise RaidError("invalid RAID level descriptor %s" % descriptor)

    def add_raid_level(self, level):
        """Adds level to levels if it is not already there.

           :param object level: an object representing a RAID level

           Raises a RaidError if level is not valid.

           Does not allow duplicate level objects.
        """
        if not self.is_raid_level(level):
            raise RaidError("level is not a valid RAID level")
        self._raid_levels.add(level)

    def __iter__(self):
        return iter(self._raid_levels)


ALL_LEVELS = RAIDLevels()


class RAID0(RAIDn):

    level = property(lambda s: "0")
    min_members = property(lambda s: 2)
    nick = property(lambda s: "stripe")

    def has_redundancy(self):
        return False

    def _get_max_spares(self, member_count):
        return 0

    def _get_base_member_size(self, size, member_count):
        return div_up(size, member_count)

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size * member_count

    def _trim(self, size, chunk_size):
        return size - size % chunk_size

    def _pad(self, size, chunk_size):
        return size + (chunk_size - (size % chunk_size)) % chunk_size

    def _get_recommended_stride(self, member_count):
        return member_count * 16


class Striped(RAID0):
    """ subclass with canonical lvm name """
    name = 'striped'
    names = [name]


RAID0 = RAID0()
ALL_LEVELS.add_raid_level(RAID0)
Striped = Striped()
ALL_LEVELS.add_raid_level(Striped)


class RAID1(RAIDn):
    level = property(lambda s: "1")
    min_members = property(lambda s: 2)
    nick = property(lambda s: "mirror")

    def has_redundancy(self):
        return True

    def _get_max_spares(self, member_count):
        return member_count - self.min_members

    def _get_base_member_size(self, size, member_count):
        return size

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size

    def _trim(self, size, chunk_size):
        return size

    def _pad(self, size, chunk_size):
        return size

    def _get_recommended_stride(self, member_count):
        return None

    def get_size(self, member_sizes, num_members=None, chunk_size=None, superblock_size_func=None):
        if not member_sizes:
            return Size(0)

        if num_members is None:
            num_members = len(member_sizes)

        min_size = min(member_sizes)
        superblock_size = superblock_size_func(min_size)
        min_data_size = self._trim(min_size - superblock_size, chunk_size)
        return self.get_net_array_size(num_members, min_data_size)


RAID1 = RAID1()
ALL_LEVELS.add_raid_level(RAID1)


class RAID4(RAIDn):
    level = property(lambda s: "4")
    min_members = property(lambda s: 3)
    nick = property(lambda s: None)

    def has_redundancy(self):
        return True

    def _get_max_spares(self, member_count):
        return member_count - self.min_members

    def _get_base_member_size(self, size, member_count):
        return div_up(size, member_count - 1)

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size * (member_count - 1)

    def _trim(self, size, chunk_size):
        return size - size % chunk_size

    def _pad(self, size, chunk_size):
        return size + (chunk_size - (size % chunk_size)) % chunk_size

    def _get_recommended_stride(self, member_count):
        return (member_count - 1) * 16


RAID4 = RAID4()
ALL_LEVELS.add_raid_level(RAID4)


class RAID5(RAIDn):
    level = property(lambda s: "5")
    min_members = property(lambda s: 3)
    nick = property(lambda s: None)

    def has_redundancy(self):
        return True

    def _get_max_spares(self, member_count):
        return member_count - self.min_members

    def _get_base_member_size(self, size, member_count):
        return div_up(size, (member_count - 1))

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size * (member_count - 1)

    def _trim(self, size, chunk_size):
        return size - size % chunk_size

    def _pad(self, size, chunk_size):
        return size + (chunk_size - (size % chunk_size)) % chunk_size

    def _get_recommended_stride(self, member_count):
        return (member_count - 1) * 16


RAID5 = RAID5()
ALL_LEVELS.add_raid_level(RAID5)


class RAID6(RAIDn):
    level = property(lambda s: "6")
    min_members = property(lambda s: 4)
    nick = property(lambda s: None)

    def has_redundancy(self):
        return True

    def _get_max_spares(self, member_count):
        return member_count - self.min_members

    def _get_base_member_size(self, size, member_count):
        return div_up(size, member_count - 2)

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size * (member_count - 2)

    def _trim(self, size, chunk_size):
        return size - size % chunk_size

    def _pad(self, size, chunk_size):
        return size + (chunk_size - (size % chunk_size)) % chunk_size

    def _get_recommended_stride(self, member_count):
        return None


RAID6 = RAID6()
ALL_LEVELS.add_raid_level(RAID6)


class RAID10(RAIDn):
    level = property(lambda s: "10")
    min_members = property(lambda s: 4)
    nick = property(lambda s: None)

    def has_redundancy(self):
        return True

    def _get_max_spares(self, member_count):
        return member_count - self.min_members

    def _get_base_member_size(self, size, member_count):
        return div_up(size, (member_count // 2))

    def _get_net_array_size(self, member_count, smallest_member_size):
        return smallest_member_size * (member_count // 2)

    def _trim(self, size, chunk_size):
        return size

    def _pad(self, size, chunk_size):
        return size + (chunk_size - (size % chunk_size)) % chunk_size

    def _get_recommended_stride(self, member_count):
        return None


RAID10 = RAID10()
ALL_LEVELS.add_raid_level(RAID10)


class Container(RAIDLevel):
    name = "container"
    names = [name]
    min_members = 1
    nick = property(lambda s: None)
    is_uniform = property(lambda s: False)

    def has_redundancy(self):
        raise RaidError("redundancy is not a concept that applies to containers")

    def get_max_spares(self, member_count):
        # pylint: disable=unused-argument
        raise RaidError("get_max_spares is not defined for level container")

    def get_space(self, size, num_members, chunk_size=None, superblock_size_func=None):
        # pylint: disable=unused-argument
        return size

    def get_recommended_stride(self, member_count):
        # pylint: disable=unused-argument
        raise RaidError("get_recommended_stride is not defined for level container")

    def get_size(self, member_sizes, num_members=None, chunk_size=None, superblock_size_func=None):
        # pylint: disable=unused-argument
        return sum(member_sizes, Size(0))


Container = Container()
ALL_LEVELS.add_raid_level(Container)


class ErsatzRAID(RAIDLevel):

    """ A superclass for a raid level which is not really a raid level at
        all, just a bunch of block devices of possibly differing sizes
        thrown together. This concept has different names depending on where
        it crops up. btrfs's name is single, lvm's is linear. Consequently,
        this abstract class implements all the functionality, but there are
        distinct subclasses which have different names.
    """
    min_members = 1
    nick = property(lambda s: None)
    is_uniform = property(lambda s: False)

    def has_redundancy(self):
        return False

    def get_max_spares(self, member_count):
        return member_count - self.min_members

    def get_space(self, size, num_members, chunk_size=None, superblock_size_func=None):
        # pylint: disable=unused-argument
        if superblock_size_func is None:
            raise RaidError("superblock_size_func value of None is not acceptable")
        return size + num_members * superblock_size_func(size)

    def get_recommended_stride(self, member_count):
        # pylint: disable=unused-argument
        return None

    def get_size(self, member_sizes, num_members=None, chunk_size=None, superblock_size_func=None):
        # pylint: disable=unused-argument
        if not member_sizes:
            return Size(0)

        if superblock_size_func is None:
            raise RaidError("superblock_size_func value of None is not acceptable")

        total_space = sum(member_sizes, Size(0))
        superblock_size = superblock_size_func(total_space)
        return total_space - len(member_sizes) * superblock_size


class Linear(ErsatzRAID):

    """ subclass with canonical lvm name """
    name = 'linear'
    names = [name, 'jbod']


Linear = Linear()
ALL_LEVELS.add_raid_level(Linear)


class Single(ErsatzRAID):

    """ subclass with canonical btrfs name. """
    name = 'single'
    names = [name]


Single = Single()
ALL_LEVELS.add_raid_level(Single)


class Dup(RAIDLevel):

    """ A RAID level which expresses one way btrfs metadata may be distributed.

        For this RAID level, duplication occurs within a single block device.
    """
    name = 'dup'
    names = [name]
    min_members = 1
    nick = property(lambda s: None)
    is_uniform = property(lambda s: False)

    def has_redundancy(self):
        return True


Dup = Dup()
ALL_LEVELS.add_raid_level(Dup)


def get_raid_level(descriptor):
    """ Convenience function to return a RAID level for the descriptor.

        :param object descriptor: a RAID level descriptor
        :rtype: RAIDLevel
        :returns: The RAIDLevel object for this descriptor

        Note that descriptor may be any object that identifies a
        RAID level, including the RAID object itself.

        Raises a RaidError is there is no RAID object for the descriptor.
    """
    return ALL_LEVELS.raid_level(descriptor)
